﻿/*	File			BullyElectionAlgorithm.cs
	Purpose			Implementation of the Bully Election Algorithm
	Author			Richard Anthony	(R.J.Anthony@gre.ac.uk)
	Date			January 2015

	Demonstration implementation of the Bully election algorithm.
	Normal process states in this algorithm are State_Slave and State_Master.
	
	All processes are initialised in State_Slave. In this state they listen for MT_MasterHeartbeat messages,
	sent periodically by a master (if one exists). If one or more slaves miss a sequence of the heartbeat 
	messages (indicating absence of a master) they initiate an election by sending a MT_ElectionTrigger message.

	The process that wins the election will send MT_MasterHeartbeat messages periodically. A very important
	rule is that there must only be one State_Master instance present at any time, and therefore the election
	mechanism must be robust to prevent multiple masters locally determining that they have won an election.
	To facilitate this a process initially enters the State_Candidate state during an election. This is a 
	transient state used to ensure that all challenges have been resolved before a single process elevates to
	State_Master.

	The following design assumptions are made:
	1.	Elections will occur between processes in a local network and therefore broadcast communication can be used.
	2.	Only one participating process will be present on each computer, so the IP address of the host computer can be
		used as the process’ ID for the purposes of the election (this is a simple way to ensure that IDs are unique).
	3.	The process with the highest ID (its IP address) wins an election.
	4.	Once a single master exists, slaves with higher IDs may join the system (without triggering an election).
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace BullyElectionAlgorithm
{
    public partial class CBullyElectionAlgorithm : Form
    {
        public const Int16 BULLY_ELECTION_ALGORITHM_PORT = 8005;	// All messages are received ion this port, regardless of the processes election state
        public const Int32 MASTER_HEARTBEAT_TX_PERIOD = 5000;       // Heartbeat transmission period (milliseconds)
        public const Int32 ELECTION_TRIGGER_HEARTBEATS_MISSED = 3;  // Threshold number of heartbeats a Slave must miss before initiating an election
        public const Int32 CHECK_FOR_MESSAGE_PERIOD = 100;          // Period between checking for any incomming messages (milliseconds)
        public const Int32 CANDIDATE_STATE_DURATION = 3;            // The number of heartbeat periods a node must wait in Candidate state before promoting
        public const Int32 HeartbeatClock_LIMIT = 1000;
        public const Int32 TIMER_MASTER_HEARTBEAT_TX = 1;
        public const Int32 TIMER_CHECK_FOR_MESSAGE = 2;

        public enum BULLY_ELECTION_ALGORITHM_Process_State {State_Slave, State_Candidate, State_Master};
        public enum BULLY_ELECTION_ALGORITHM_Message_Type {MT_MasterHeartbeat, MT_ElectionTrigger};

        // Message structure (this will be serialised before sending)
        // Serialisation / deserialisation is very simple in this case because there is only a single field with a fixed size
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct BULLY_ELECTION_ALGORITHM_Message
        {
	        public BULLY_ELECTION_ALGORITHM_Message_Type eMessageType;
        };

        Int32 m_iBULLY_ELECTION_ALGORITHM_Message_Size;

	    BULLY_ELECTION_ALGORITHM_Process_State m_LocalState;

        Socket m_SendSock;
        Socket m_RecvSock;
        IPEndPoint m_SendEndPoint;      // Address for sending all messages (all are broadcast)
        IPEndPoint m_LocalEndPoint;     // Local Address for receiving all messages
        System.Windows.Forms.Timer m_MasterHeartbeat_Timer;
        System.Windows.Forms.Timer m_MessageReceive_Timer;

	    Int32 m_iMasterHeartbeatClock;	// Used to measure time intervals between heartbeats from the Master instance
	    Int32 m_iMostRecentHeartbeat_TimeDetected;
	    Int32 m_iClock_CountInCandidateState_SincelastChallenge; // Keeps track of how long the node has been in Candidate state
						    // Must stay in candidate state long enough to ensure all possible challengers have been defeated

        public CBullyElectionAlgorithm()
        {
            InitializeComponent();
            m_iBULLY_ELECTION_ALGORITHM_Message_Size = sizeof(BULLY_ELECTION_ALGORITHM_Message_Type);
 	        m_LocalState = BULLY_ELECTION_ALGORITHM_Process_State.State_Slave;
	        m_iMasterHeartbeatClock = 0;	// Used to measure time intervals between heartbeats from the Master instance
	        m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock;
	        m_iClock_CountInCandidateState_SincelastChallenge = 0;
            GetLocalHostAddressDetails();
	        SetUpSendAddressStructForBroadcast();   // All election algorithm messages will be broadcast
            CreateSendSocket();                     // The socket used for all sends
	        CreateAndBindReceiveSocket();	        // The socket on which all messages are received
           
 	        InitialiseControls();
	        Start_MasterHeartbeatTimer();
	        Start_MessageReceiveTimer();
	        WriteEventListEntry("Initialised (Slave state)");
        }

        void InitialiseControls()
        {
 	        DisplayLocalState();
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {
                    m_LocalEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], BULLY_ELECTION_ALGORITHM_PORT);
                    return;
                }
            }
            // No IPv4 address found - local host will not be able to connect to server using IPv4
            MessageBox.Show("Failed to get local address!","BULLY_ELECTION_ALGORITHM");
            CloseSocketAndExit();
        }

        void SetUpSendAddressStructForBroadcast()
        {
            m_SendEndPoint = new IPEndPoint(IPAddress.Broadcast, BULLY_ELECTION_ALGORITHM_PORT);
        }

       void CreateSendSocket()
        {
            try
            {
                m_SendSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_SendSock.Blocking = true;         // Set blocking mode
                m_SendSock.EnableBroadcast = true;  // Set broadcast mode
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Bully Election Algorithm");
                CloseSocketAndExit();
            }
        }

        void CreateAndBindReceiveSocket()
        {   
            try
            {
                m_RecvSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_RecvSock.Blocking = false; // Set non-blocking mode
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Bully Election Algorithm");
                CloseSocketAndExit();
            }

            try
            {
                m_RecvSock.Bind(m_LocalEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use", "Bully Election Algorithm");
                CloseSocketAndExit();
            }
        }

        void Start_MasterHeartbeatTimer()
        {	
            m_MasterHeartbeat_Timer = new System.Windows.Forms.Timer();
            m_MasterHeartbeat_Timer.Interval = MASTER_HEARTBEAT_TX_PERIOD; // milliseconds
            m_MasterHeartbeat_Timer.Tick += OnTimedEvent_MasterHeartbeatTimer;
            m_MasterHeartbeat_Timer.Enabled = true;
        }

        void Stop_MasterHeartbeatTimer()
        {
            if (null != m_MasterHeartbeat_Timer)
            {
                m_MasterHeartbeat_Timer.Enabled = false;
            }
        }

        void Start_MessageReceiveTimer()
        {	
            m_MessageReceive_Timer = new System.Windows.Forms.Timer();
            m_MessageReceive_Timer.Interval = CHECK_FOR_MESSAGE_PERIOD; // milliseconds
            m_MessageReceive_Timer.Tick += OnTimedEvent_MessageReceiveTimer;
            m_MessageReceive_Timer.Enabled = true;
        }

        void Stop_MessageReceiveTimer()
        {
            if (null != m_MessageReceive_Timer)
            {
                m_MessageReceive_Timer.Enabled = false;
            }
        }

        void OnTimedEvent_MasterHeartbeatTimer(Object myObject, EventArgs myEventArgs)
        {
		    m_iMasterHeartbeatClock++;		// Used to measure time intervals between heartbeats from the Master (only relevent when in Slave state)
		    m_iMasterHeartbeatClock %= HeartbeatClock_LIMIT;	// m_iHeartbeatClock counts from 0 to 999, then is reset to 0 and start counting again
		
		    switch(m_LocalState) {
			    case BULLY_ELECTION_ALGORITHM_Process_State.State_Master:
				    SendMasterHeartbeatMessage();	// Only performed if current state is State_Master	
				    break;
			    case BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate:
				    m_iClock_CountInCandidateState_SincelastChallenge++; // Increment Candidate duration clock
				    if(CANDIDATE_STATE_DURATION <= m_iClock_CountInCandidateState_SincelastChallenge)
				    {	// Have stayed in candidate state for the requisite period (no successful challengers, so now promote to Master)				
					    TransitionToMasterState();
				    }
				    break;
			    case BULLY_ELECTION_ALGORITHM_Process_State.State_Slave:
				    CheckPresenceOfMaster(m_iMasterHeartbeatClock);		// Check master is still present based on interval since last heartbeat received
				    break;
            }
		}

        void OnTimedEvent_MessageReceiveTimer(Object myObject, EventArgs myEventArgs)
        {
		    DoReceive();
        }

        void CheckPresenceOfMaster(int iCurrentHeartbeatClock)
        {	// Check interval since last heartbeat to determine if a Master instance is still present	
	        int iEffective_CurrentHeartbeatClock; // Takes care of clock rollovers
	
	        // Heartbeat count increments every MASTER_HEARTBEAT_TX_PERIOD milliseconds, up to a maximum of HeartbeatClock_LIMIT increments (modulo)
	        // Heartbeat count rolls over when it reaches HeartbeatClock_LIMIT, back to 0 (%HeartbeatClock_LIMIT)
	        if(iCurrentHeartbeatClock < m_iMostRecentHeartbeat_TimeDetected)
	        {	// Current clock value is lower, so it has rolled over! Treat it as having a value of +HeartbeatClock_LIMIT
		        // for the purposes of the calculations (so the current clock value is always higher than the stored value
		        // in iMostRecentHeartbeat_TimeDetected).
		        iEffective_CurrentHeartbeatClock = iCurrentHeartbeatClock + HeartbeatClock_LIMIT;
	        }
	        else
	        {	// Use Current clock value as is.
		        iEffective_CurrentHeartbeatClock = iCurrentHeartbeatClock;
	        }

	        if((m_iMostRecentHeartbeat_TimeDetected + ELECTION_TRIGGER_HEARTBEATS_MISSED)
		        <= iEffective_CurrentHeartbeatClock)
	        {	// m_iMostRecentHeartbeat_TimeDetected value is out-of-date, Master is considered to have departed
		        TransitionToCandidateState(); //Start election, with self as a candidate
	        }
        }

        void TransitionToMasterState()
        {
	        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Master);
	        SendMasterHeartbeatMessage();	// Immediately send a heartbeat message to prevent further slave elections
        }

        void TransitionToCandidateState()
        {
	        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate);
	        m_iClock_CountInCandidateState_SincelastChallenge = 0; // Reset Candidate duration clock
	        SendElectionTriggerMessage();	//Start election, with self as a candidate
        }

        void TransitionToSlaveState()
        {
	        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Slave);
	        m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock; 	// Update MostRecentHeartbeat timestamp
        }

        byte[] Serialize_Message_PDU(BULLY_ELECTION_ALGORITHM_Message Message)
        {   // Convert an Message_PDU structure into a byte array
            // Serialisation is very simple in this case, since there is only one Field in the message structure
            byte[] bByteArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            return bByteArray;
        }

        bool SendMessage(BULLY_ELECTION_ALGORITHM_Message Message)
        {
            byte[] bBuffer = Serialize_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_SendSock.SendTo(bBuffer, m_SendEndPoint);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
            }
            return true;
        }

        void SendMasterHeartbeatMessage()
        {
            BULLY_ELECTION_ALGORITHM_Message Message;
            Message.eMessageType = BULLY_ELECTION_ALGORITHM_Message_Type.MT_MasterHeartbeat;
            SendMessage(Message);
            WriteEventListEntry("MasterHeartbeatMessage sent");
        }

        void SendElectionTriggerMessage()
        {
            BULLY_ELECTION_ALGORITHM_Message Message;
            Message.eMessageType = BULLY_ELECTION_ALGORITHM_Message_Type.MT_ElectionTrigger;
            SendMessage(Message);
            WriteEventListEntry("ElectionTriggerMessage sent");
        }

        BULLY_ELECTION_ALGORITHM_Message DeSerialize_ReceivedMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            // Serialisation is very simple in this case, since there is only one Field in the message structure
            BULLY_ELECTION_ALGORITHM_Message result_Message = new BULLY_ELECTION_ALGORITHM_Message();
            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (BULLY_ELECTION_ALGORITHM_Message_Type)ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            return result_Message;
        }

        void DoReceive()
        {
            byte[] bArray = new byte[m_iBULLY_ELECTION_ALGORITHM_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            IPAddress SenderIPAddress;
            try
            {
                iReceiveByteCount = m_RecvSock.ReceiveFrom(bArray, ref From_EndPoint);
                SenderIPAddress = ((IPEndPoint)From_EndPoint).Address; // Store sender's IPAddress for comparison purposes
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                return;
            }

            if (0 >= iReceiveByteCount)
            {
                return;
            }

            if (true == CompareNodeIDs_SenderSameAsLocal(SenderIPAddress))
            {	// Ignore messages from self (very important when using broadcast)
                WriteEventListEntry("Ignore message from self");
                return;
            }

            BULLY_ELECTION_ALGORITHM_Message Message = DeSerialize_ReceivedMessage(bArray);

            switch (Message.eMessageType)
            {
                case BULLY_ELECTION_ALGORITHM_Message_Type.MT_MasterHeartbeat:
                    WriteEventListEntry("MasterHeartbeatMessage received");
                    switch (m_LocalState)
                    {
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Master: // Another Master exists (the one with lowest IP address should demote to Slave state)
                            WriteEventListEntry("M recd M_Heartbeat");
                            CompareNodeIDs_DemoteIfRemoteIsHigher(SenderIPAddress);
                            break;
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate: // A Master exists (abandon election and demote to Slave state)
                            WriteEventListEntry("C recd M_Heartbeat");
                            TransitionToSlaveState();
                            break;
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Slave: // A Master exists (no action needed)
                            WriteEventListEntry("S recd M_Heartbeat");
                            m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock; 	// Update MostRecentHeartbeat timestamp
                            break;
                    }
                    break;
                case BULLY_ELECTION_ALGORITHM_Message_Type.MT_ElectionTrigger:
                    WriteEventListEntry("ElectionTriggerMessage received");
                    switch (m_LocalState)
                    {
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Master: // Another node thinks the Master does not exist - send heartbeat immediately
                            WriteEventListEntry("M recd Election");
                            SendMasterHeartbeatMessage();
                            break;
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate: // This is the 'election proper' (the candidate with lowest IP address should demote to Slave state)
                            WriteEventListEntry("C recd Election");
                            CompareNodeIDs_DemoteIfRemoteIsHigher(SenderIPAddress);
                            if (BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate == m_LocalState)
                            {	// Still in Candidate state, so won the challenge
                                SendElectionTriggerMessage(); // Signal the local node's ID to other nodes (do this every time we are challenged)
                                m_iClock_CountInCandidateState_SincelastChallenge = 0; // Reset Candidate duration clock
                                WriteEventListEntry("Local won");
                            }
                            break;
                        case BULLY_ELECTION_ALGORITHM_Process_State.State_Slave:	// This is the 'election proper' (if the local ID is higher than the sender's then we join the election)
                            WriteEventListEntry("S recd Election");
                            CompareNodeIDs_PromoteToCandidateIfRemoteIsLower(SenderIPAddress);
                            if (BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate == m_LocalState)
                            {	// Now in Candidate state, so won the challenge
                                WriteEventListEntry("Local won");
                            }
                            break;
                    }
                    break;
            }
        }

        void CompareNodeIDs_DemoteIfRemoteIsHigher(IPAddress SenderIPAddress)
        {
            if (false == CompareNodeIDs_RemoteIsLower(SenderIPAddress))
	        {	// Remote has higher address - the local node looses and must demote to Slave state
		        TransitionToSlaveState();
	        }
        }

        void CompareNodeIDs_PromoteToCandidateIfRemoteIsLower(IPAddress SenderIPAddress)
        {
            if (true == CompareNodeIDs_RemoteIsLower(SenderIPAddress))
	        {	// Remote has lower address - the local node wins and must promote to Candidate state
		        TransitionToCandidateState();
	        }
        }

        bool CompareNodeIDs_RemoteIsLower(IPAddress SenderIPAddress)
        {
            byte[] bLocalAd = m_LocalEndPoint.Address.GetAddressBytes();
            UInt32 iLocal = (UInt32) ((bLocalAd[0]*256*256*256) + (bLocalAd[1]*256*256) + (bLocalAd[2]*256) + bLocalAd[3]);
            byte[] bSenderAd = SenderIPAddress.GetAddressBytes();
            UInt32 iSender = (UInt32) ((bSenderAd[0]*256*256*256) + (bSenderAd[1]*256*256) + (bSenderAd[2]*256) + bSenderAd[3]);
            if (iLocal > iSender)
            {	// Local node has the higher ID (its IP address)
                return true;
            }
            // Note, since only one process can exist at each computer (due to port binding), the case where both addresses are equal cannot occur.
 	        return false; // Local node has the lower ID (its IP address)
        }

        bool CompareNodeIDs_SenderSameAsLocal(IPAddress SenderIPAddress)
        {
            return m_LocalEndPoint.Address.Equals(SenderIPAddress);
        }

        void ChangeState(BULLY_ELECTION_ALGORITHM_Process_State NewState)
        {
            string sPrevious = ConvertLocalStateToString(); // Remember previous state for display update
	        m_LocalState = NewState;                    // Do the state update
            string sNew = ConvertLocalStateToString();	
	        string sDisplay;
	        sDisplay = string.Format("StateChange {0} -> {1}",sPrevious,sNew);
	        WriteEventListEntry(sDisplay);
	        DisplayLocalState();
        }

        void DisplayLocalState()
        {
            CurrentState_textBox.Text = ConvertLocalStateToString();
        }

        string ConvertLocalStateToString()
        {
            switch (m_LocalState)
            {
                case BULLY_ELECTION_ALGORITHM_Process_State.State_Master:
                    return "Master";
                case BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate:
                    return "Candidate";
                case BULLY_ELECTION_ALGORITHM_Process_State.State_Slave:
                    return "Slave";
            }
            return "Unknown";
        }

        void WriteEventListEntry(string sLine)
        {
            Event_listBox.Items.Add(sLine);
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        private void ClearEventList_button_Click(object sender, EventArgs e)
        {
            Event_listBox.Items.Clear();
        }

        private void Done_button_Click_1(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            Stop_MasterHeartbeatTimer();
            Stop_MessageReceiveTimer();
            try
            {
                m_RecvSock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_SendSock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            Application.Exit();
        }
    }
}
