﻿/*	File			NoughtsCrossesCLIENT.cs
	Purpose			Network Game (Noughts and Crosses)CLIENT
	Author			Richard John Anthony
	Date			January 2015 (original Cpp version December 2003)
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace CaseStudyGame_Client
{
    public partial class NoughtsCrossesCLIENT : Form
    {
        public const Int16 PORT = 8000;     // Default port used by the Game application 8000 decimal
        public const uint From_DIRECTORY_SERVICE_PORT = 8002;	// Fixed, defined by Directory Service
        public const uint To_DIRECTORY_SERVICE_PORT = 8008;	    // Fixed, defined by Directory Service

        public const Int32 MAX_ALIAS_LEN = 15;

        public enum MessageType {REGISTER_ALIAS, PLAYER_LIST, CHOOSE_OPPONENT, LOCAL_CELL_SELECTION,
					    OPPONENT_CELL_SELECTION, START_OF_GAME, END_OF_GAME, CLIENT_DISCONNECT};

        public enum CodeType {END_OF_GAME_DRAW, END_OF_GAME_WIN, END_OF_GAME_LOOSE};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Message_PDU
        {
	        public MessageType eMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ALIAS_LEN + 1))]
	        public byte[] cAlias;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = ((MAX_ALIAS_LEN + 1)*10))]
	        public byte[] cPlayerList;      // Comma-separated list of aliases
            public Int32 iCell;
            public CodeType eCode;
            public Int32 iToken;
            public bool bFirstMove;
        };

        public const int DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH = 50;
        public const string GAME_SERVER_NAME = "GameServer";	// Game Server name, as registered with DS

        // Structure format of Directory Service Request messages (sent to DS)
        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceRequest_PDU
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH)]
            public byte[] bBuffer;
        };

        // Structure format of Directory Service Reply messages (received from DS)
        [StructLayout(LayoutKind.Sequential)]
        public struct DirectoryServiceReply_PDU  // Fixed, defined by Directory Service
        {   // IP address followed by port
            public byte a1;
            public byte a2;
            public byte a3;
            public byte a4;
            public uint iPort;
        };

        Socket m_ResolveSock;
        Socket m_GameConnectSocket;
        IPEndPoint m_DirectoryService_IPEndPoint;
        IPEndPoint m_localIPEndPoint;
        IPEndPoint m_GameConnectEndPoint;
        System.Windows.Forms.Timer m_DirectoryServiceReply_Timer;
        System.Windows.Forms.Timer m_ApplicationMessage_Timer;

        // Helper variables used in message Serialize, Deserialize and Receive
        Int32 m_iFieldSize1; // eMessageType
        Int32 m_iFieldSize2; // cAlias
        Int32 m_iFieldSize3; // cPlayerList
        Int32 m_iFieldSize4; // iCell
        Int32 m_iFieldSize5; // eCode
        Int32 m_iFieldSize6; // iToken
        Int32 m_iFieldSize7; // bFirstMove
        Int32 m_iPadSize_To4ByteBoundary; // Size of pad field to be added to meet a 4-byte boundary for interoperability with C++
        Int32 m_iMessageSize; // The structure itself, plus any padding

       	public bool m_bConnected;
        string m_sToken;
	    string m_sOpponentToken;
  
        public NoughtsCrossesCLIENT()
        {
            InitializeComponent();
            m_bConnected = false;
            m_iMessageSize = CalculateMessageStructureSizeAndFieldSizes(ref m_iFieldSize1, ref m_iFieldSize2, ref m_iFieldSize3,
                        ref m_iFieldSize4, ref m_iFieldSize5, ref m_iFieldSize6, ref m_iFieldSize7, ref m_iPadSize_To4ByteBoundary);
            InitialiseControls();

            if (true == CreateResolveSocket())
            {
                bool bResult = BindResolveSocket();
                if (true == bResult)
                {
                    ContactDirectoryService();
                }
            }
        }

        Int32 CalculateMessageStructureSizeAndFieldSizes(ref Int32 F1, ref Int32 F2, ref Int32 F3,
                             ref Int32 F4, ref Int32 F5, ref Int32 F6, ref Int32 F7, ref Int32 Pad)
        {  // Helper method, relating to the use of the Message_PDU message structure
            // simplifies message Serialize, Deserialize and Receive
            F1 = sizeof(MessageType);           // eMessageType
            F2 = MAX_ALIAS_LEN + 1;             // cAlias
            F3 = (MAX_ALIAS_LEN + 1) * 10;      // cPlayerList
            F4 = sizeof(Int32);                 // iCell
            F5 = sizeof(CodeType);              // eCode
            F6 = sizeof(Int32);                 // iToken
            F7 = sizeof(bool);                  // bFirstMove
            Int32 iContentMessageSize = F1 + F2 + F3 + F4 + F5 + F6 + F7;
            Pad = CalculatePadTo4ByteBoundary(iContentMessageSize);
            Int32 iPaddedMessageSize = iContentMessageSize + Pad;
            return iPaddedMessageSize;
        }

        Int32 CalculatePadTo4ByteBoundary(Int32 iContentMessageSize)
        {   // Helper method, relating to the use of the Message_PDU message structure
            // calculates the size of the padding neccesary to truncate the message on a 4-byte boundary (compatible with C++)
            // Pad will always be between 0 and 3 bytes
            int iRemainder = iContentMessageSize % 4; // The extent of overlap beyond 4-byte boundary
            if (0 == iRemainder)
            {   // Already on a 4-byte boundary
                return 0; // No padding necessary
            }
            else
            {   // Calculate pad to next 4-byte boundary
                return 4 - iRemainder;
            }
        }
           
        void InitialiseControls()
        {
            //Port_textBox.Text = Convert.ToString(PORT);
            Port_textBox.Enabled = false;
            ConnectAddress_1_TextBox.Enabled = false;
            ConnectAddress_2_TextBox.Enabled = false;
            ConnectAddress_3_TextBox.Enabled = false;
            ConnectAddress_4_TextBox.Enabled = false;
            ContactWithDS_textBox.Enabled = false;
            label_Connected.Visible = false;
            label_NotConnected.Visible = true;
            RegisterAlias_Button.Enabled = false;
            AliasName_TextBox.Enabled = false;
            SelectPlayer_Button.Enabled = false;
            PlayerListBox.Enabled = false;
            GameStatus_TextBox.Text = "No active game"; // Display game status
            Token_TextBox.Text = "";
            MoveStatus_TextBox.Text = "";
            OpponentAlias_TextBox.Text = "";
            OpponentAlias_TextBox.Enabled = false;
            ClearAllCells();
        }

        void CreateGameSocket()
        {
            try
            {
                m_GameConnectSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_GameConnectSocket.Blocking = true; // Set blocking mode for socket during connection set-up phase
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "'Noughts and Crosses' Client");
                CloseSocketAndExit();
            }
        }

        void ConnectToGameServer(DirectoryServiceReply_PDU DirectoryServiceReply_message)
        {
            CreateGameSocket();

            // Setup Application Address (to game server, based on reply from DS)
        	// The Game Server address and port is retrieved from the Directory Service
            byte[] bAddress = new byte[4];
            bAddress[0] = DirectoryServiceReply_message.a1;
            bAddress[1] = DirectoryServiceReply_message.a2;
            bAddress[2] = DirectoryServiceReply_message.a3;
            bAddress[3] = DirectoryServiceReply_message.a4;
            int iPort = System.Convert.ToInt16(DirectoryServiceReply_message.iPort);
            
            // Display the Game server address
            ConnectAddress_1_TextBox.Text = bAddress[0].ToString();
            ConnectAddress_2_TextBox.Text = bAddress[1].ToString();
            ConnectAddress_3_TextBox.Text = bAddress[2].ToString();
            ConnectAddress_4_TextBox.Text = bAddress[3].ToString();
            Port_textBox.Text = iPort.ToString();

            try
            {
                IPAddress ipHostAddress = new IPAddress(bAddress);
                m_GameConnectEndPoint = new System.Net.IPEndPoint(ipHostAddress, iPort);
                m_GameConnectSocket.Connect(m_GameConnectEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {   // Connect attempt was NOT successful, alert user and return
                MessageBox.Show("Connect() failed - is server running?, are port and IP address values correct?", 
                                "'Noughts and Crosses' Client");
                return;
            }
	
	        m_bConnected = true;
            label_Connected.Visible = true;
            label_NotConnected.Visible = false;
            AliasName_TextBox.Enabled = true;	
	        RegisterAlias_Button.Enabled = true;

	        // Set non-blocking (now that connection established) for receiving
            try
            {
                m_GameConnectSocket.Blocking = false; // Set non-blocking mode
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Could not set non-Blocking behaviour for socket", "'Noughts and Crosses' Client");
            }

            Start_Application_Timer(); // Enable receiving (check periodically)
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private byte[] Serialize_Message_PDU(Message_PDU Message)
        {   // Convert an Message_PDU structure into a byte array
            byte[] bByteArray = new byte[m_iMessageSize];        // Create a byte array to hold the serialised form of the Message PDU

            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize1/*count*/);

            iFieldOffset += m_iFieldSize1;                  // Adjust offset for second field
            int iDataLength2 = Message.cAlias.Length;       // Get the length of data actually stored in field2 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cAlias/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength2/*count*/);

            iFieldOffset += m_iFieldSize2;                  // Adjust offset for third field
            int iDataLength3 = Message.cPlayerList.Length;  // Get the length of data actually stored in field3 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cPlayerList/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength3/*count*/);

            iFieldOffset += m_iFieldSize3;                  // Adjust offset for fourth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCell);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize4/*count*/);

            iFieldOffset += m_iFieldSize4;                  // Adjust offset for fifth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.eCode);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize5/*count*/);

            iFieldOffset += m_iFieldSize5;                  // Adjust offset for sixth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iToken);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize6/*count*/);

            iFieldOffset += m_iFieldSize6;                  // Adjust offset for seventh field
            bWorkingArray = BitConverter.GetBytes((bool)Message.bFirstMove); // Boolean is 1 byte
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize7/*count*/);

            // Encoding the padding (to ensure message size is aligned with a 4-byte boundary): The bByteArray has been created to the full message
            // size including the pad. This byte array is automatically initialised such that all bytes contain zeros, therefore no additional
            // encoding is necessary to include the padding at the end of the message. The padding is always between 0 and 3 bytes.
            return bByteArray;
        }

        bool SendMessage(Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_GameConnectSocket.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_GameConnectSocket))
                    {
                    }
                    else // still connected
                    {
                        return false;
                    }
                }
            }
            return true;
         }
        
        void Start_Application_Timer()
        {
            m_ApplicationMessage_Timer = new System.Windows.Forms.Timer();
            m_ApplicationMessage_Timer.Interval = 100; // milliseconds
            m_ApplicationMessage_Timer.Tick += OnTimedEvent_Application_Timer;
            m_ApplicationMessage_Timer.Enabled = true;
        }

        void Stop_Application_Timer()
        {
            if (null != m_ApplicationMessage_Timer)
            {
                m_ApplicationMessage_Timer.Enabled = false;
            }
        }

        void OnTimedEvent_Application_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                DoReceive();
            }
            catch // Silently catch any errors
            {
            }
        }

        void DoReceive()
        {
            byte[] bArray = new byte[m_iMessageSize];
            int iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_GameConnectSocket.Receive(bArray);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_GameConnectSocket))
                    {
                    }
                    else // still connected
                    {
                        return;
                    }
                }
            }
 
 		    if(0 == iReceiveByteCount)
		    {
			    return;
		    }

            Message_PDU Message = DeSerialize_ReceivedMessage(bArray);

		    switch(Message.eMessageType)
		    {
                case MessageType.PLAYER_LIST:
		            DisplayPlayerList(Message);	
		            break;
                case MessageType.START_OF_GAME:
			        SelectPlayer_Button.Enabled = false;            // Disable Select Opponent control
			        string sOpponent = "";
                    int iLength = ByteArrayTo_ASCII_String(Message.cAlias, ref sOpponent);	// Display opponent alias
                    OpponentAlias_TextBox.Text = sOpponent;
                    
                    if(Message.iToken == 0) 						// Set and display our token
			        {
				        m_sToken = "0";
				        m_sOpponentToken = "X";
			        }
			        else
			        {
				        m_sToken = "X";
				        m_sOpponentToken = "0";
			        }		
			        Token_TextBox.Text = m_sToken;
			
			        if(Message.bFirstMove == true)					// configure first-move logic
			        {
				        MoveStatus_TextBox.Text = "Your Move";
				        EnableEmptyCells();
			        }
			        else
			        {
				        MoveStatus_TextBox.Text = "Opponent's Move";
				        DisableEmptyCells();
			        }
			        GameStatus_TextBox.Text = "Game In Progress"; // Display game status

 		            break;
                case MessageType.END_OF_GAME:
			        switch(Message.eCode) 
			        {
                        case CodeType.END_OF_GAME_DRAW:
			                MessageBox.Show("The game was a draw","Game Over");	
			                break;
                        case CodeType.END_OF_GAME_WIN:
                            MessageBox.Show("Well done! You won", "Game Over");	
			                break;
                        case CodeType.END_OF_GAME_LOOSE:
                            MessageBox.Show("Unlucky - You lost", "Game Over");	
			                break;
			        }
                    CloseSocketAndExit();
			        break;
                case MessageType.OPPONENT_CELL_SELECTION:
		            switch (Message.iCell)
			        {
		            case 0:
                        Cell0_Button.Text = m_sOpponentToken;
                        Cell0_Button.Enabled = false;
                        break;
		            case 1:
                        Cell1_Button.Text = m_sOpponentToken;
                        Cell1_Button.Enabled = false;
                        break;
		            case 2:
                        Cell2_Button.Text = m_sOpponentToken;
                        Cell2_Button.Enabled = false;
                        break;
		            case 3:
                        Cell3_Button.Text = m_sOpponentToken;
                        Cell3_Button.Enabled = false;
                        break;
		            case 4:
                        Cell4_Button.Text = m_sOpponentToken;
                        Cell4_Button.Enabled = false;
                        break;
		            case 5:
                        Cell5_Button.Text = m_sOpponentToken;
                        Cell5_Button.Enabled = false;
                        break;
		            case 6:
                        Cell6_Button.Text = m_sOpponentToken;
                        Cell6_Button.Enabled = false;
                        break;
		            case 7:
                        Cell7_Button.Text = m_sOpponentToken;
                        Cell7_Button.Enabled = false;
                        break;
		            case 8:
                        Cell8_Button.Text = m_sOpponentToken;
                        Cell8_Button.Enabled = false;
                        break;
			        }
			    MoveStatus_TextBox.Text = "Your Move";
			    EnableEmptyCells();
			    break;
                default:
                break;
		    }
	    }

        public Message_PDU DeSerialize_ReceivedMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            Message_PDU result_Message = Message_PDU_Initialise();
            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);

            iFieldOffset += m_iFieldSize1;
            result_Message.cAlias = new byte[MAX_ALIAS_LEN + 1];
            for (int iIndex = 0; iIndex < m_iFieldSize2; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cAlias[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize2;
            result_Message.cPlayerList = new byte[(MAX_ALIAS_LEN + 1) * 10];
            for (int iIndex = 0; iIndex < m_iFieldSize3; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cPlayerList[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize3;
            result_Message.iCell = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize4;
            result_Message.eCode = (CodeType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize5;
            result_Message.iToken = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize6;
            result_Message.bFirstMove = BitConverter.ToBoolean(bRecdByteArray, iFieldOffset); // Converts a single byte within a byte array to a bool value   

            // Messages are natively aligned on 4-byte boundaries by C++ send(). To ensure compatibility, C# messages are adjusted in the
            // application code to truncate on 4-byte boundaries. The padding is added at the end of the message, so any bytes beyond the last
            // expected field can be ignored when receiving, i.e. no special decoding is needed for the padding when a message is DeSerializeed
            return result_Message;
        }

        void DisplayPlayerList(Message_PDU Message)
        {
	        PlayerListBox.Items.Clear();            // Clear contents of PlayerList control
	        string sAlias = AliasName_TextBox.Text; // Read our Alias (exclude it from displayed list)
            string sPlayerList = "";
            int ilength = ByteArrayTo_ASCII_String(Message.cPlayerList, ref sPlayerList);
            // sPlayerList is a comma-separated, ASCII (single byte per char) null-terminated list (terminated at iLength)
	        // There is no comma after the last item in the list
            int iNextCommaPos;
	        string sCurrentPlayer = "";
            while ((iNextCommaPos = sPlayerList.IndexOf(',')) > 0 && iNextCommaPos < ilength) // Ignore chars beyond the valid string content
            {
                sCurrentPlayer = sPlayerList.Substring(0/*start*/, iNextCommaPos/*length*/);
                if (!sCurrentPlayer.Equals(sAlias, StringComparison.OrdinalIgnoreCase)) // Exclude own alias from displayed list
                {
                    PlayerListBox.Items.Add(sCurrentPlayer);
                }
                    sPlayerList = sPlayerList.Substring(iNextCommaPos + 1/*start*/);// Remove 'dealt with' name from head of list
                    ilength -= (iNextCommaPos + 1);                                 // Keep track of remaining list-string length
	        }
            // Deal with residual substring (there is no comma after the last item in the list)
            sPlayerList = TruncateString(sPlayerList, ilength);
            if (0 < ilength && !sPlayerList.Equals(sAlias, StringComparison.OrdinalIgnoreCase)) // Exclude own alias from displayed list
            {
                 PlayerListBox.Items.Add(sPlayerList);
            }

            if (0 >= PlayerListBox.Items.Count)
            {
                PlayerListBox.Enabled = false;
                SelectPlayer_Button.Enabled = false;
            }
            else
            {
                PlayerListBox.SelectedIndex = 0; // Set default selection
                PlayerListBox.Enabled = true;
                SelectPlayer_Button.Enabled = true;
            }
        }

        void ConnectionLost()
        {
            CloseSocketAndExit();
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        private void RegisterAlias_Button_Click(object sender, EventArgs e) // Register Alias with server
        {
	        string sAlias = AliasName_TextBox.Text;
	        if(0 == sAlias.Length)
	        {
                MessageBox.Show("Alias name cannot be blank", "'Noughts and Crosses' Client");
	        }
	        else
	        {
                Message_PDU Message = Message_PDU_Initialise();
                Message.eMessageType = MessageType.REGISTER_ALIAS;
                sAlias = TruncateString(sAlias, MAX_ALIAS_LEN); // Truncate if necessary
                AliasName_TextBox.Text = sAlias; // Update textbox with (possibly) truncated string
                Message.cAlias = StringTo_ASCII_ByteArray(sAlias);
                if(SendMessage(Message) == true)
		        {
			        AliasName_TextBox.Enabled = false;	
			        RegisterAlias_Button.Enabled = false;
		        }
	        }
        }

        private void SelectPlayer_Button_Click(object sender, EventArgs e)
        {
            int iSelection = PlayerListBox.SelectedIndex;
	        if(-1 == iSelection)
	        {
                MessageBox.Show("No player selected", "'Noughts and Crosses' Client");
	        }
	        else
	        {
		        // Send a CHOOSE_OPPONENT message to server
                Message_PDU Message = Message_PDU_Initialise(); 
                Message.eMessageType = MessageType.CHOOSE_OPPONENT;
                Message.cAlias = StringTo_ASCII_ByteArray(PlayerListBox.SelectedItem.ToString());
                SendMessage(Message);
	        }
        }

       private void Cell0_Button_Click(object sender, EventArgs e)
        {
 	        Cell0_Button.Text = m_sToken;
	        Cell0_Button.Enabled = false;
	        CellSelectAction(0);
        }

        private void Cell1_Button_Click(object sender, EventArgs e)
        {
 	        Cell1_Button.Text = m_sToken;
	        Cell1_Button.Enabled = false;
	        CellSelectAction(1);
        }

        private void Cell2_Button_Click(object sender, EventArgs e)
        {
 	        Cell2_Button.Text = m_sToken;
	        Cell2_Button.Enabled = false;
	        CellSelectAction(2);
        }

        private void Cell3_Button_Click(object sender, EventArgs e)
        {
  	        Cell3_Button.Text = m_sToken;
	        Cell3_Button.Enabled = false;
	        CellSelectAction(3);
        }

        private void Cell4_Button_Click(object sender, EventArgs e)
        {
 	        Cell4_Button.Text = m_sToken;
	        Cell4_Button.Enabled = false;
	        CellSelectAction(4);
        }

        private void Cell5_Button_Click(object sender, EventArgs e)
        {
  	        Cell5_Button.Text = m_sToken;
	        Cell5_Button.Enabled = false;
	        CellSelectAction(5);
        }

        private void Cell6_Button_Click(object sender, EventArgs e)
        {
 	        Cell6_Button.Text = m_sToken;
	        Cell6_Button.Enabled = false;
	        CellSelectAction(6);
        }

        private void Cell7_Button_Click(object sender, EventArgs e)
        {
 	        Cell7_Button.Text = m_sToken;
	        Cell7_Button.Enabled = false;
	        CellSelectAction(7);
        }

        private void Cell8_Button_Click(object sender, EventArgs e)
        {
 	        Cell8_Button.Text = m_sToken;
	        Cell8_Button.Enabled = false;
	        CellSelectAction(8);
        }

        void CellSelectAction(int iCell)
        {   // send a CELL_SELECTION message to server
            Message_PDU Message = Message_PDU_Initialise(); 
            Message.eMessageType = MessageType.LOCAL_CELL_SELECTION;
	        Message.iCell = iCell;
            if(SendMessage(Message) == true)
	        {
                MoveStatus_TextBox.Text = "Opponent's Move";
		        DisableEmptyCells();
	        }
        }

        void DisableEmptyCells()
        {
	        if("" == Cell0_Button.Text) Cell0_Button.Enabled = false; 
	        if("" == Cell1_Button.Text) Cell1_Button.Enabled = false; 
	        if("" == Cell2_Button.Text) Cell2_Button.Enabled = false; 
	        if("" == Cell3_Button.Text) Cell3_Button.Enabled = false; 
	        if("" == Cell4_Button.Text) Cell4_Button.Enabled = false; 
	        if("" == Cell5_Button.Text) Cell5_Button.Enabled = false; 
	        if("" == Cell6_Button.Text) Cell6_Button.Enabled = false; 
	        if("" == Cell7_Button.Text) Cell7_Button.Enabled = false; 
	        if("" == Cell8_Button.Text) Cell8_Button.Enabled = false; 
        }

        void EnableEmptyCells()
        {
	        if("" == Cell0_Button.Text) Cell0_Button.Enabled = true; 
	        if("" == Cell1_Button.Text) Cell1_Button.Enabled = true; 
	        if("" == Cell2_Button.Text) Cell2_Button.Enabled = true; 
	        if("" == Cell3_Button.Text) Cell3_Button.Enabled = true; 
	        if("" == Cell4_Button.Text) Cell4_Button.Enabled = true; 
	        if("" == Cell5_Button.Text) Cell5_Button.Enabled = true; 
	        if("" == Cell6_Button.Text) Cell6_Button.Enabled = true; 
	        if("" == Cell7_Button.Text) Cell7_Button.Enabled = true; 
	        if("" == Cell8_Button.Text) Cell8_Button.Enabled = true; 
        }

        void ClearAllCells()
        {
            Cell0_Button.Text = "";
            Cell1_Button.Text = "";
            Cell2_Button.Text = "";
            Cell3_Button.Text = "";
            Cell4_Button.Text = "";
            Cell5_Button.Text = "";
            Cell6_Button.Text = "";
            Cell7_Button.Text = "";
            Cell8_Button.Text = "";
	        DisableEmptyCells();
        }        
        
        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            Application.Exit();
        }

        void CloseConnectionAndSocket()
        {
            Stop_Application_Timer();
            try
            {
                m_GameConnectSocket.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
            try
            {
                m_GameConnectSocket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        Message_PDU Message_PDU_Initialise()
        {
            Message_PDU M = new Message_PDU();
            M.eMessageType = MessageType.END_OF_GAME; // This field is always overwritten as used in every instance
            M.cAlias = StringTo_ASCII_ByteArray("");
            M.cPlayerList = StringTo_ASCII_ByteArray("");
            M.iCell = 0;
            M.eCode = CodeType.END_OF_GAME_DRAW;
            M.iToken = 0;
            M.bFirstMove = false;
            return M;
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            //Encoding.ASCII.GetBytes(sStr);
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }

        private void ContactDirectoryService()
        {    // Initialise timer to wait for directory service reply (uses non-blocking socket)
            m_DirectoryServiceReply_Timer = new System.Windows.Forms.Timer();
            m_DirectoryServiceReply_Timer.Interval = 250; // 250 millisecond period
            m_DirectoryServiceReply_Timer.Tick += new EventHandler(On_DirectoryServiceReply_Timer);
            m_DirectoryServiceReply_Timer.Enabled = true;
            SendResolveMessage();
        }

        bool CreateResolveSocket()
        {   // UDP, non-blocking, broadcast
            try
            {
                m_ResolveSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_ResolveSock.Blocking = false;
                m_ResolveSock.EnableBroadcast = true;
                System.Net.IPAddress BroadcastIPAddress = System.Net.IPAddress.Broadcast;
                m_DirectoryService_IPEndPoint = new System.Net.IPEndPoint(BroadcastIPAddress, System.Convert.ToInt16(To_DIRECTORY_SERVICE_PORT));
                return true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Create Socket failed");
                return false;
            }
            catch   // Catch any other exceptions (e.g. arising from creating the Endpoint)
            {
                MessageBox.Show("Invalid IP address or Port number");
                return false;
            }
        }

        void SendResolveMessage() // Broadcast message to Directory Server
        {
            String sRequest = "RESOLVE:" + GAME_SERVER_NAME;
            // Convert the message string into a byte array format (!C# chars are two bytes each!)
            byte[] bBuffer = Encoding.ASCII.GetBytes(sRequest);
            m_ResolveSock.SendTo(bBuffer, m_DirectoryService_IPEndPoint);
            ContactWithDS_textBox.Text = "Resolve request sent to DS";
        }

        public bool BindResolveSocket()
        {
            // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(From_DIRECTORY_SERVICE_PORT));
            try
            {   // Bind to the local IP Address and selected port
                m_ResolveSock.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
                MessageBox.Show("bind failed");
                return false;
            }
        }

        private void On_DirectoryServiceReply_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                Receive_DirectoryServiceReply();
            }
            catch // Catch any errors
            {
            }
        }

        private void Receive_DirectoryServiceReply()
        {
            EndPoint localEndPoint = (EndPoint)m_localIPEndPoint;  // New EndPoint structure to hold the address of the sender 
            byte[] bBuffer = new byte[DIRECTORY_SERVICE_MAX_REQUEST_MESSAGE_LENGTH];
            int iReceiveByteCount;
            iReceiveByteCount = m_ResolveSock.ReceiveFrom(bBuffer, ref localEndPoint);
            if (0 < iReceiveByteCount)
            {   // Stop the timer now the Directory Service reply has been received
                m_DirectoryServiceReply_Timer.Enabled = false;
                // Close the resolve socket, so the Directory Service port is released and thus other Directory Service clients can run 
                // on the same computer (this is very useful for development and testing purposes)
                m_ResolveSock.Close();
                // DeSerialise the message received                
                DirectoryServiceReply_PDU DirectoryServiceReply_message = DeSerialize_DirectoryServiceReply(bBuffer);
                if (DirectoryServiceReply_message.iPort != 0)	// Port of 0 means App Service was not known
                {   // Use the resolved Game Server address information to create a TCP connection with the Game server
                    if (false == m_bConnected)
                    {
                        ContactWithDS_textBox.Text = "Reply received from DS";
                        ConnectToGameServer(DirectoryServiceReply_message);
                    }
                }
            }
        }

        public DirectoryServiceReply_PDU DeSerialize_DirectoryServiceReply(byte[] received)
        {   // Converts a byte array to a DirectoryServiceReply_PDU structure (all fields fixed data length)
            IntPtr ip = Marshal.AllocHGlobal(received.Length);  // Allocate unmanaged memory (working storage)
            Marshal.Copy(received, 0, ip, received.Length);     //Copy the byte array to the working storage
            // Marshal the unmanaged memory contents to the DirectoryServiceReply_PDU structure
            DirectoryServiceReply_PDU structure = (DirectoryServiceReply_PDU)Marshal.PtrToStructure(ip, typeof(DirectoryServiceReply_PDU));
            Marshal.FreeHGlobal(ip);                            // Free working storage
            return structure;
        }
    }
}
